/*
 * Copyright (c) 1991, 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
char copyright[] =
    "@(#) Copyright (c) 1991, 1992 The Regents of the University of California.\nAll rights reserved.\n";
static  char rcsid[] =
    "@(#)$Header: dexpire.c,v 1.74 92/01/26 13:31:56 leres Exp $ (LBL)";
#endif

/*
 * dexpire - dynamic expire for netnews
 *
 * Most (if not all) other expire programs force the news administrator
 * to predict how many days worth of news articles they have room for.
 * This approach is fundamentally incorrect.
 *
 * A better way is to specify the relative importance or priority
 * of newsgroups and have the expire program delete just enough
 * articles in the desired proportion to free the required number
 * of disk blocks. This is how dexpire works.
 *
 * Dexpire breaks all newsgroups into "classes." Each class has a
 * priority associated with it. High priority groups are kept for
 * a longer period than low priority groups.
 *
 * The "standard" class is the highest priority class. Articles in
 * newsgroups in the standard class are kept the longest. The standard
 * class is also used to determine how long lower priority classes
 * are kept.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef SYSV
#include <strings.h>
#else
#include <string.h>
#define rindex strrchr
#endif
#include <ctype.h>
#include <signal.h>

#include "dexpire.h"
#include "disk.h"
#include "file.h"
#include "util.h"
#include "patchlevel.h"
#ifdef EXPIRES_DBZ
#include "dbz.h"
#endif

/* Class node */
struct class {
	long pri;		/* priority of this class */
	long deleted;		/* number of articles deleted */
	long freed_blk;		/* number of 512 bytes blocks freed */
	time_t minimum;		/* minimum time to keep articles */
	time_t maximum;		/* maximum time to keep articles */
	time_t to;		/* timestamp of oldest article in class */
	struct group *group;	/* group list */
	struct class *next;	/* forward link */
};

/* Expire rule list */
struct rule {
	char *ngn;		/* newsgroup name, e.g. alt/slack */
	char mod;		/* moderation character */
	char chud[3];
	struct class *class;	/* pointer back to class */
	struct rule *next;	/* forward link */
};

/* Newsgroup node */
struct group {
	char *ngn;		/* newsgroup name, e.g. alt/slack */
	long no;		/* number of oldest article */
	long nn;		/* number of newest article */
	time_t to;		/* timestamp of oldest article */
#ifdef USE_SCANDIR
	int *nums;              /* article numbers array */
	int nnums;              /* size of aforementioned item */
	int ino;                /* running counter, index in nums */
#endif
	struct group *next;	/* forward link */
};

/* Converts 512 byte blocks to Kbytes */
#define BLK2KB(b) ((b) / 2)

#ifdef SYSV
#define B2BLK(b) (((b) + 511) / 512)
#endif

/* Private data */
static char *active_file;
static char *spool_dir;
static char *dev_dir;
static char *dexplist;
static char *dfeedback;
static char *ndfeedback;
#ifdef EXPIRES_DBZ
static char *expires;
#endif
static dev_t spool_dev;

static long free_kb;		/* current number of free Kbytes in spool_dir */
static long used_kb;		/* current number of used Kbytes in spool_dir */
static long free_ino;		/* current number of free inodes in spool_dir */
static long inodes;		/* current number of used inodes in spool_dir */
static long want_kb = 4000;	/* desired number of free Kbytes in spool_dir */
static long want_ino = 1000;	/* desired number of free inodes in spool_dir */
static long need_kb;		/* needed number of free Kbytes in spool_dir */
static long need_ino;		/* needed number of free Kbytes in spool_dir */
static long freed_blk;		/* number of 512 bytes blocks freed */

static time_t standardtime;	/* "standard" timestamp */
static time_t currenttime;	/* current timestamp */
static time_t deltatime = 0;	/* delta of current and standard timestamps */
static time_t dtime;		/* delta added to standardtime in loop */
static time_t ioldesttime;	/* initial oldest timestamp */

static int deleted;		/* number of articles deleted */
static double density;		/* Kbytes/inode for spool dir */
static togo;			/* iterations before dtime calculation */

#ifdef USE_SCANDIR
static int dirlists = 0;        /* groups processed normally */
static int dirscans = 0;        /* groups that needed scandir */
extern int scannumdir();
#endif

static struct class *classlist = 0;
static struct class *standardclass = 0;
static struct rule *rulelist = 0;

static FILE *feed;		/* feedback file descriptor */

static int daemonsecs = 0;      /* daemon mode flag and check interval */
static time_t daemonrestart;    /* daemon mode restart time */
static char *pidfile=NULL;      /* file where we record our PID */

static int newgroup(), nextrule();
static int expireone(), dexpire();
static void findstandard(), newdtime(), dump(),
       pfcleanup(), sighand(), set_signals();
#define enable_signals() set_signals(sighand);
#define disable_signals() set_signals(SIG_IGN);


/* Public data */
char *prog;

int nflag = 0;			/* non-destructive mode */
int feedback = 0;		/* update the feedback file */
int debug = 0;			/* debugging modes */
int verbose = 0;		/* verbose information to stdout */
#ifdef EXPIRES_DBZ
int checkexp = 0;               /* check the explicit expires db */
#endif
#ifdef S_ISLNK
int zombies = 0;                /* dangling symlinks count */
#endif

/* External data */

extern char *optarg;
extern int optind, opterr;

extern char *version;

int
main(argc, argv)
	int argc;
	char **argv;
{
	register int op, status=0;
	register char *cp;
	register FILE *f;
	register struct class *cl;
	register struct group *gl;
	register int i;
	time_t d, oldesttime;
	double r1, r2;
	struct stat sbuf;
	int firsttime = 1;
	char *usage =
   "usage: %s [-dnuv] [-a active] [-c dexplist] [-s spool_dir] [-S check_dir]\n [-f Kbytes] [-i inodes] [-D wait,restart] [-P pidfile]\n";

	if ((cp = rindex(argv[0], '/')))
		prog = cp + 1;
	else
		prog = argv[0];

	setvbuf(stdout, NULL, _IOLBF, 0);
	active_file=savestr(ctlfile("active"));
	dev_dir=spool_dir=savestr(fullartfile(NULL));
	dexplist=savestr(ctlfile("dexplist"));
	dfeedback=savestr(ctlfile("dfeedback"));
	ndfeedback=savestr(ctlfile("dfeedback.new"));
#ifdef EXPIRES_DBZ
	expires=savestr(ctlfile("expires"));
#endif
	/* Process arguments */
	while ((op = getopt(argc, argv, "dnuva:c:f:i:s:S:D:P:")) != EOF)
		switch (op) {

		case 'a':
			active_file = optarg;
			break;

		case 'c':
			dexplist = optarg;
			break;

		case 'd':
			++debug;
			break;

		case 'f':
			want_kb = atoi(optarg);
			if (want_kb <= 0) {
				(void)fprintf(stderr,
				    "%s: \"%s\" invalid argument to -f\n",
				    prog, optarg);
				exit(1);
				/* NOTREACHED */
			}
			/* Trailing 'm' means Mbytes */
			for (cp = optarg; isdigit(*cp); ++cp)
				;
			if (*cp == 'm' || *cp == 'M')
				want_kb *= 1024;
			break;

		case 'i':
			want_ino = atoi(optarg);
			if (want_ino <= 0) {
				(void)fprintf(stderr,
				    "%s: \"%s\" invalid argument to -i\n",
				    prog, optarg);
				exit(1);
				/* NOTREACHED */
			}
			break;

		case 'n':
			++nflag;
			break;

		case 's':
			spool_dir = optarg;
			break;

		case 'S':
			dev_dir = optarg;
			break;

		case 'u':
			++feedback;
			break;

		case 'v':
			++verbose;
			break;

		case 'D':
			daemonsecs = atoi(optarg);
			for (cp = optarg; isdigit(*cp); ++cp)
				;
			if (*cp == ',')
				daemonrestart = atoi(++cp);
			else
				daemonrestart = 60*60*24; /* one day */
			if ((daemonsecs<=0) || (daemonrestart<=0)) {
				(void)fprintf(stderr,
				    "%s: \"%s\" invalid argument to -D\n",
				    prog, optarg);
				exit(1);
				/* NOTREACHED */
			}
			daemonrestart += time(0);
			break;

		case 'P':
			pidfile = optarg;
			break;

		default:
			(void)fprintf(stderr, usage, prog);
			exit(1);
			/* NOTREACHED */
		}

	if (optind != argc) {
		(void) fprintf(stderr, usage, prog);
		exit(1);
		/* NOTREACHED */
	}
	if (verbose) {
		(void)printf("%s: Version %s, patchlevel %d\n",
		    prog, version, PATCHLEVEL);
		(void)printf("%s: Want %ld Kbyte%s, %ld inode%s free\n",
		    prog, want_kb, PLURAL(want_kb), want_ino, PLURAL(want_ino));
		(void)printf("%s: Spool directory is %s\n", prog, spool_dir);
		(void)printf("%s: Working on device of %s\n", prog, dev_dir);
	}
	if (pidfile) {
	    FILE *pf = fopen(pidfile, "w");
	    if (pf) {
		(void)fprintf(pf, "%d\n%s\n", getpid(), dev_dir);
		(void)fclose(pf);
		(void)atexit(pfcleanup);
	    } else {
		(void)fprintf(stderr, "%s: warning: can't record PID\n");
	    }
	}
	enable_signals();

    repeat:
	freed_blk = 0;
	deleted = 0;
	/* Fetch current time which is used in various calculations */
	currenttime = time(0);
	if (verbose) {
		(void)printf("%s: Current time: %s\n",
		    prog, fmtdate(currenttime));
	}

	/* Update various disk parameters */
	if (disk_usage(dev_dir, &used_kb, &free_kb, &inodes, &free_ino) < 0) {
		(void)fprintf(stderr, "%s: disk_usage(): ", prog);
		perror(dev_dir);
		exit(1);
		/* NOTREACHED */
	}
	if (stat(dev_dir, &sbuf) < 0) {
		(void)fprintf(stderr, "%s: stat(): ", prog);
		perror(dev_dir);
		exit(1);
		/* NOTREACHED */
	}
	spool_dev = sbuf.st_dev;

	/* Calculate how much space we need */
	need_kb = want_kb - free_kb;
	if (free_ino<0)
		need_ino = -1;
	else
		need_ino = want_ino - free_ino;
	if (verbose)
		(void)printf("%s: Have %ld Kbyte%s, %ld inode%s free\n",
		    prog, free_kb, PLURAL(free_kb), free_ino, PLURAL(free_ino));

	/* Bail if we already have enough free space */
	if ((need_kb <= 0) && (need_ino <=0)) {
		(void)printf("%s: Nothing to do!\n", prog);
		goto noneed;
	}
	if (verbose && (need_kb>0))
		(void)printf("%s: Need to free %ld Kbyte%s\n",
		    prog, need_kb, PLURAL(need_kb));
	if (verbose && (need_ino>0))
		(void)printf("%s: Need to free %ld inode%s\n",
		    prog, need_ino, PLURAL(need_ino));

	/* Complain if asked to free too much */
	r1 = ((double)want_kb) / (((double) free_kb) + ((double)used_kb));
	r2 = ((double)want_ino) / (((double) free_ino) + ((double)inodes));
	if (((r1 >= MAX_FREE) || (r2 >= MAX_FREE)) && !nflag) {
		(void)fprintf(stderr,
		    "%s: Cannot free %d%% of the filesystem (%d%% max)",
		    prog, (int)(100.0 * (r1>r2?r1:r2)), (int)(100.0 * MAX_FREE));
		(void)fprintf(stderr, "; use -n if testing\n");
		exit(1);
		/* NOTREACHED */
	}

	/* Calculate file system density (in Kbytes/inode) */
	if (verbose)
		(void)printf("%s: Spool directory ", prog);
	if (inodes < 0) {
		density = DEFAULT_DENSITY;
		if (verbose)
			(void)printf("inode usage unknown, assuming ");
	} else {
		density = ((double)used_kb) / ((double)inodes);
		if (verbose)
			(void)printf("averages ");
	}
	if (verbose) {
		(void)printf("%.1f Kbytes/inode\n", density);
	}
#ifdef EXPIRES_DBZ
	if (dbminit(expires) == 0) {
	        dbzincore(1);
		checkexp = 1;
	} else {
                (void)fprintf(stderr, "%s: warning: ", prog);
		perror(expires);
                (void)fprintf(stderr, "%s: disregarding Expires lines\n", prog);
		/* but continue work */
	}
#endif

	if (firsttime) {
		/* Read in the dexplist config file */
		if ((f = fopen(dexplist, "r")) == 0) {
		    (void)fprintf(stderr, "%s: fopen(): ", prog);
		    perror(dexplist);
		    exit(1);
		    /* NOTREACHED */
		}
		if (!readexplist(f, nextrule)) {
		    exit(1);
		    /* NOTREACHED */
		}
		(void)fclose(f);

		/* Read in the active file */
		if ((f = fopen(active_file, "r")) == 0) {
		    (void)fprintf(stderr, "%s: fopen(): ", prog);
		    perror(active_file);
		    exit(1);
		    /* NOTREACHED */
		}
		if (!readactive(f, newgroup)) {
		    exit(1);
		    /* NOTREACHED */
		}
		(void)fclose(f);

		/* Dump out data structures if debugging */
		if (debug>1)
		    dump();

		/* Determine the "standard" time */
		findstandard();

		firsttime = 0;
	}

	/* Create the new feedback file, if necessary */
	if (feedback && !nflag) {
		if (verbose)
			(void)printf("%s: Creating feedback file\n", prog);
		if ((feed = fopen(ndfeedback, "w")) == 0) {
			(void)fprintf(stderr, "%s: fopen(): ", prog);
			perror(ndfeedback);
			status |= 1;
			feedback = 0;
		}
	}

	/* Go delete some articles */
	status = dexpire();

#ifdef EXPIRES_DBZ
	if (checkexp)
		dbmclose();
#endif

	/* Update the feedback file, if necessary */
	if (feedback) {
		if (nflag) {
			(void)fprintf(stderr,
			    "%s: Skipping feedback file update\n", prog);
		} else {
			if (verbose)
				(void)printf("%s: Updating feedback file\n",
				    prog);
			for (cl = classlist; cl; cl = cl->next)
				for (gl = cl->group; gl; gl = gl->next)
					(void)fprintf(feed, "%ld\t%s\n",
					    gl->to, gl->ngn);
			(void)fclose(feed);
			disable_signals();
			if (rename(ndfeedback, dfeedback) < 0 &&
			    unlink(dfeedback) < 0 &&
			    rename(ndfeedback, dfeedback) < 0) {
				(void)fprintf(stderr, "%s: rename(): %s -> ",
				    prog, ndfeedback);
				perror(dfeedback);
				status |= 1;
			}
			enable_signals();
		}
	}

	if (verbose) {
		/* Report final class times */
		for (cl = classlist; cl; cl = cl->next)
			(void)printf("%s: Final %3d%%: %s (%s)\n",
			    prog, (100 * cl->pri) / standardclass->pri,
			    fmtdate(cl->to), fmtdelta(currenttime - cl->to));

		/* Report per-class deletion statistics */
		for (cl = classlist; cl; cl = cl->next) {
			(void)printf("%s: Class %3d%%",
			    prog, (100 * cl->pri) / standardclass->pri);
			if (cl->minimum) {
				i = cl->minimum / (24 * 60 * 60);
				(void)printf(" (%d day%s)", i, PLURAL(i));
			}
			(void)printf(": %seleted %ld article%s (%ld Kbyte%s)\n",
			    nflag ? "Would have d" : "D",
			    cl->deleted, PLURAL(cl->deleted),
			    BLK2KB(cl->freed_blk),
			    PLURAL(BLK2KB(cl->freed_blk)));
		}

		(void)printf(
		    "%s: %seleted %d article%s total (%ld Kbyte%s), took %s\n",
		    prog, nflag ? "Would have d" : "D",
		    deleted, PLURAL(deleted),
		    BLK2KB(freed_blk), PLURAL(BLK2KB(freed_blk)),
		    fmtdelta(time(0) - currenttime));
#ifdef S_ISLNK
		(void)printf("%s: Deleted %d dead symlink%s\n",
                    prog, zombies, PLURAL(zombies));
#endif

		/* Determine new oldest article time */
		oldesttime = currenttime;
		for (cl = classlist; cl; cl = cl->next)
		    if (!cl->minimum)
			for (gl = cl->group; gl; gl = gl->next)
			    if (gl->to < oldesttime)
				oldesttime = gl->to;

		d = currenttime - ioldesttime;
		(void)printf("%s: Old oldest with no minimum %s (%s)\n",
		    prog, fmtdate(ioldesttime), fmtdelta(d));
		d = currenttime - oldesttime;
		(void)printf("%s: New oldest with no minimum %s (%s)\n",
		    prog, fmtdate(oldesttime), fmtdelta(d));

		d = ioldesttime - oldesttime;
		cp = "gained";
		if (d < 0) {
			d = -d;
			cp = "lost";
		}
		(void)printf("%s: Article delta %s %s\n",
		    prog, cp, fmtdelta(d));
		ioldesttime = oldesttime;
	}

    noneed:
	if (daemonsecs) {
		sleep(daemonsecs);
		if (time(0) > daemonrestart) {
		    execv(argv[0], argv);
		    (void)fprintf(stderr, "%s: Can't restart daemon\n", prog);
		    exit(99);
		    /* NOTREACHED */
		}
		goto repeat;
	}

	exit(status);
	/* NOTREACHED */
}

/* Handle the next rule */
static int
nextrule(group, mod, pri, mindays, maxdays)
	char *group, mod;
	int pri, mindays, maxdays;
{
	register struct class *cl, *cl2;
	register struct rule *rp;
	register time_t minimum, maximum;
	static struct rule *lastrule = 0;

	if (pri < 0) {
		(void)fprintf(stderr,
		    "%s: priority for group %s is negative (%d)",
		    prog, group, pri);
		exit(1);
		/* NOTREACHED */
	}
	if (mindays < 0) {
		(void)fprintf(stderr,
		    "%s: minimum days for group %s is negative (%d)",
		    prog, group, mindays);
		exit(1);
		/* NOTREACHED */
	}
	minimum = mindays * 24 * 60 * 60;
	if (maxdays < 0) {
		(void)fprintf(stderr,
		    "%s: maximum days for group %s is negative (%d)",
		    prog, group, maxdays);
		exit(1);
		/* NOTREACHED */
	}
	maximum = maxdays * 24 * 60 * 60;

	/* Save new rule */
	rp = (struct rule *)mymalloc(sizeof(*rp), "rule");
	bzero((char *)rp, sizeof(*rp));
	if (rulelist == 0)
		rulelist = rp;
	else
		lastrule->next = rp;
	lastrule = rp;
	rp->ngn = savestr(group);
	rp->mod = mod;

	/* Don't create a new class for priority 0 */
	if (pri == 0)
		return(1);

	/* Look for existing class with the same priority and minimum */
	cl2 = 0;
	for (cl = classlist; cl; cl = cl->next) {
		if (cl->pri == pri && cl->minimum == minimum && cl->maximum == maximum)
			break;
		if (cl->pri < pri || (cl->pri == pri && cl->minimum < minimum))
			cl2 = cl;
	}

	/* New class; insert in ascending order */
	if (cl == 0) {
		cl = (struct class *)mymalloc(sizeof(*cl), "class");
		bzero((char *)cl, sizeof(*cl));
		if (cl2) {
			cl->next = cl2->next;
			cl2->next = cl;
		} else {
			cl->next = classlist;
			classlist = cl;
		}
		cl->pri = pri;
		cl->minimum = minimum;
		cl->maximum = maximum;
	}

	/* Find the "standard" class */
	for (cl2 = classlist; cl2; cl2 = cl2->next)
		standardclass = cl2;

	/* Backpointer from new rule to class */
	rp->class = cl;

	return(1);
}

/* Handle the next newsgroup from the active file */
static int
newgroup(group, nn, no, mod)
	char *group;
	int nn, no;
	char mod;
{
	register struct group *gl;
	register struct class *cl;
	register struct rule *rp;
	register char *cp, *cp2;

	/* Locate the class for this newsgroup */
	for (rp = rulelist; rp; rp = rp->next) {
		/* Check moderated flag */
		if (rp->mod != mod &&
		    rp->mod != 'x' &&
		    (rp->mod != 'u' || mod != 'y'))
			continue;

		/* Check for a group match */
		cp = group;
		cp2 = rp->ngn;
		while (*cp != '\0' && *cp == *cp2) {
			++cp;
			++cp2;
		}
		if (*cp2 == '\0' && (*cp == '\0' || *cp == '/'))
			break;

		/* "all" always matches */
		if (strcmp(rp->ngn, "all") == 0)
			break;
	}

	/* Gotta have a rule */
	if (rp == 0) {
		(void)fprintf(stderr,
		    "%s: Warning: no default rule, discarding %s\n",
		    prog, group);
		return(1);
	}

	/* Ingore groups in the 0 priority class */
	if (rp->class == 0) {
		if (verbose)
			(void)printf("%s: Skipping %s\n", prog, group);
		if (!nflag && feedback)
			(void)fprintf(feed, "0\t%s\n", group);
		return(1);
	}

	/* Build a new group entry */
	gl = (struct group *)mymalloc(sizeof(*gl), "group");
	bzero((char *)gl, sizeof(*gl));

	gl->ngn = savestr(group);
	gl->no = no;
	gl->nn = nn;
#ifdef USE_SCANDIR
	gl->nums = NULL;
	gl->nnums = 0;
	gl->ino = 0;
#endif
	cl = rp->class;
	gl->next = cl->group;
	cl->group = gl;

	return(1);
}

/* Fetch initial timestamp info */
static void
findstandard()
{
	register struct class *cl;
	register struct group *gl;
	register time_t start, t, d;
	register long n;
	register int numstats;
	register char *art;
	char file[512];
	struct stat sbuf;
#ifdef EXPIRES_DBZ
	datum dbkey, dbval;
#endif
#ifdef USE_SCANDIR
	int *nums;
	int n0;
#endif

#ifdef EXPIRES_DBZ
	/* Disregard explicit expires */
#define CHECKEXP \
	  if (checkexp) { (void)sprintf(file, "%s/%s", gl->ngn, art);\
	      dbkey.dptr = file; dbkey.dsize = strlen(file)+1;\
	      dbval = dbzfetch(dbkey);\
  	      if (dbval.dptr != NULL) {\
		if (verbose>1) (void)printf("%s: skipping %s\n", prog, file);\
	      continue; } }
#else
#define CHECKEXP ;
#endif

	/* Get initial timestamp in case we need it later */
	start = time(0);

	/* Determine timestamp of oldest article in each class */
	numstats = 0;
	ioldesttime = currenttime;
	for (cl = classlist; cl; cl = cl->next) {
		/* Safe initial value */
		cl->to = currenttime - 24*3600;

		for (gl = cl->group; gl; gl = gl->next) {
			/*
			 * Change our working directory. Although this
			 * seems inefficient, it's about as expensive
			 * as checking to see if the newsgroup
 exists and can greatly reduce
			 * overhead if the third field of the active
			 * file isn't up to date.
			 */
			(void)sprintf(file, "%s/%s", spool_dir, gl->ngn);
			if (debug>1) {
				(void) printf("%s\n", file);
			}
			if (chdir(file) < 0) {
				gl->to = currenttime;
				continue;
			}

			t = 0;
			art = "?";
#ifdef USE_SCANDIR
			++dirlists;
#ifdef SCANDIR_MAX
			n0 = SCANDIR_MAX;
#else
			n0 = (gl->nn - gl->no) * SCANDIR_THRES;
			if (n0 < SCANDIR_MIN)
				n0 = SCANDIR_MIN;
#endif
#endif
			for (n = gl->no; n <= gl->nn; ++n) {
#ifdef USE_SCANDIR
				if (--n0 < 0) {
				/* Read the whole directory */
					if (debug)
					  printf("  invoking scandir on %s\n",
						 gl->ngn);
					++dirscans;
					n0 = scannumdir(".", &nums);
					if (n0<0) {
					  (void)fprintf(stderr,
					   "%s: failed to read dir: %s, %d\n",
					   prog, gl->ngn, n0);
					  break;
					}
					for (n=0; n<n0; ++n) {
					  art=long2str(nums[n]);
					  ++numstats;
					  if (xstat(gl->ngn, art, &sbuf)>=0) {
					    CHECKEXP;
					    t = sbuf.st_mtime;
					    break;
					  }
					}
					gl->nums = nums;
					gl->nnums = n0;
					goto tfound;
			        }
#endif			  
				/* Construct article filename */
				art = long2str(n);

				/* Attempt to get timestamp */
				++numstats;
				if (xstat(gl->ngn, art, &sbuf)>=0) {
				  CHECKEXP;
				  t = sbuf.st_mtime;
				  goto tfound;
				}
			}
		tfound:
			if (t == 0) {
				gl->to = currenttime;
				continue;
			}

			/* Complain if from the first year of Unix history */
			if (t < 365 * 24 * 60 * 60)
				(void)fprintf(stderr,
				    "%s: Warning: %s/%s is old: %s\n",
				    prog, gl->ngn, art, fmtdate(t));

			gl->no = n;
			gl->to = t;
			if (t < cl->to) {
				cl->to = t;
				if ((t < ioldesttime) && !cl->minimum)
					ioldesttime = t;
			}
		}
	}

	/* Determine standard time */
	standardtime = currenttime;
	for (cl = classlist; cl; cl = cl->next) {
		if (verbose)
			(void)printf("%s: Oldest %3d%%: ",
			    prog, (100 * cl->pri) / standardclass->pri);

		/* Safety check */
		if (cl->to == 0) {
			if (verbose)
				(void)printf("has zero time!\n");
			continue;
		}

		/* Determine class delta time */
		d = currenttime - cl->to;
		if (verbose)
			(void)printf("%s (delta %s)\n",
			    fmtdate(cl->to), fmtdelta(d));

		/* Increase class delta by class ratio */
		d = (d * standardclass->pri) / cl->pri;

		/* Convert to absolute time */
		t = currenttime - d;
		if (verbose > 1)
			(void)printf("%s: Expanded     %s (delta %s)\n",
			    prog, fmtdate(t), fmtdelta(d));

		if (t < standardtime)
			standardtime = t;
	}

	/* Calculate delta time */
	deltatime = currenttime - standardtime;

	if (verbose) {
		(void)printf("%s: Initial Standard: %s (%s)\n",
		    prog, fmtdate(standardtime), fmtdelta(deltatime));
#ifdef USE_SCANDIR
		(void)printf("%s: listed %d, scanned %d groups\n",
			     prog, dirlists-dirscans, dirscans);
#endif
		(void)printf("%s: (Took %s to check %d file%s)\n",
		    prog, fmtdelta(time(0) - start),
		    numstats, PLURAL(numstats));
	}
}

/* Calculates new dtime and togo */
static void
newdtime()
{
	register long n, need;
	double r;

	need = need_kb - BLK2KB(freed_blk);
	if (verbose > 1)
		(void)printf("%s: newdtime(): Need to free %ld Kbyte%s\n",
		    prog, need, PLURAL(need));

	/* Calculate new togo based on file system density */
	togo = (int)(need / density * 0.25);
	if (verbose > 1)
		(void)printf("%s: New togo is ", prog);
	if (togo < MIN_TOGO) {
		if (verbose > 1)
			(void)printf("too small (%d), using ", togo);
		togo = MIN_TOGO;
	} else if (togo > MAX_TOGO) {
		if (verbose > 1)
			(void)printf("too big (%d), using ", togo);
		togo = MAX_TOGO;
	}
	if (verbose > 1)
		(void)printf("%d article%s\n", togo, PLURAL(togo));

	/* Calculate new dtime based on file system usage */
	n = used_kb;
	if (n == 0)
		n = 1;
	r = ((double)need) / ((double)n);
	if (verbose > 1)
		(void)printf("%s: Need to used ratio is %.1e\n", prog, r);
	dtime = (int)(deltatime * r * 0.25);
	if (verbose > 1)
		(void)printf("%s: New dtime is ", prog);
	if (dtime < MIN_DTIME) {
		if (verbose > 1)
			(void)printf("too short (%s), using ", fmtdelta(dtime));
		dtime = MIN_DTIME;
	} else if (dtime > MAX_DTIME) {
		if (verbose > 1)
			(void)printf("too long (%s), using ", fmtdelta(dtime));
		dtime = MAX_DTIME;
	}
	if (verbose > 1) {
		(void)printf("%s\n", fmtdelta(dtime));
	}
}

/* Make a pass over one newsgroup. Returns true if enough space was freed */
static int
expireone(gl, cl, cutoff, cutoffmax)
	register struct group *gl;
	register struct class *cl;
	register time_t cutoff, cutoffmax;
{
	register long n;
	register int i, status;
	register int olddeleted;
	register time_t t;
	register char *art;
	char file[512];
	struct stat sbuf;
#ifdef EXPIRES_DBZ
	datum dbkey, dbval;
#endif

	/* Make sure there's something to do */
	if (gl->no > gl->nn) {
		if (verbose > 2)
			(void)printf("%s: No articles in %s\n", prog, gl->ngn);
		return(0);
	}

	/* Cached article time might tell us there's nothing to do */
	/* (This test fails if time isn't cached) */
	if (gl->to >= cutoff) {
		if (verbose > 2)
			(void)printf("%s: Cached time skip of %s\n",
			    prog, gl->ngn);
		return(0);
	}

	/* Change working directory to speed up directory lookups */
	(void)sprintf(file, "%s/%s", spool_dir, gl->ngn);
	if (stat(file, &sbuf) < 0)
	        goto nodir;
	if (sbuf.st_dev != spool_dev) {
                /* On the wrong device, skip */
		if (verbose > 1)
			(void)printf("%s: Wrong device %s\n", prog, gl->ngn);
		return(0);
	}
	
	if (chdir(file) < 0) {
		/* Skip next time and save the unsuccessful chdir() */
	     nodir:
		gl->no = gl->nn + 1;
		gl->to = currenttime;
		if (verbose > 1)
			(void)printf("%s: No directory %s\n", prog, gl->ngn);
		return(0);
	}

	/* Nuke 'em, Dano! */
	if (verbose > 1)
		(void)printf("%s: Running %s\n", prog, gl->ngn);

	status = 0;
	t = 0;
	olddeleted = deleted;
	n = gl->no;
	while (n <= gl->nn) {
		/* Construct article filename */
		art = long2str(n);

		t = 0;
		/* Does this article exist? */
		if (xstat(gl->ngn, art, &sbuf) < 0)
		  goto nextart;

		/* Skip if on the wrong filesystem or no file */
		if ((sbuf.st_dev != spool_dev) || (!S_ISREG(sbuf.st_mode)))
		  goto nextart;

		/* Complain if from the first year of Unix history */
		t = sbuf.st_mtime;
		if (t < 365 * 24 * 60 * 60)
			(void)fprintf(stderr,
			    "%s: Warning: %s/%s is old: %s\n",
			    prog, gl->ngn, art, fmtdate(t));

		/* If not old enough, we're done with this group */
		if (t >= cutoff)
			break;

#ifdef EXPIRES_DBZ
		if (checkexp && (t>=cutoffmax)) {
		        /* Does this article have an explicit expire? */
		        (void)sprintf(file, "%s/%s", gl->ngn, art);
		        dbkey.dptr = file;
			dbkey.dsize = strlen(file)+1;
			dbval = dbzfetch(dbkey);
			if (dbval.dptr != NULL) {
			      if (verbose>1)
				    (void)printf("%s: skipping %s\n",
						 prog, file);
			      goto nextart;
		        }
		}
#endif
		if (!nflag) {
			/* Only chalk up the article if it was the last link */
			if (unlink(art) >= 0 && sbuf.st_nlink <= 1) {
				++deleted;
#ifndef SYSV
				freed_blk += sbuf.st_blocks;
#else
				freed_blk += B2BLK(sbuf.st_size);
#endif
				++cl->deleted;
#ifndef SYSV
				cl->freed_blk += sbuf.st_blocks;
#else
				cl->freed_blk += B2BLK(sbuf.st_size);
#endif
			}
		} else {
			/* Pretend to delete the article */
			++deleted;
			freed_blk += sbuf.st_blocks;
			++cl->deleted;
			cl->freed_blk += sbuf.st_blocks;
		}

		if ((BLK2KB(freed_blk) >= need_kb) && (deleted>=need_ino)) {
			if (verbose)
				(void)printf("%s: We're done!\n", prog);
			status = 1;
			break;
		}

		/* Occasionally recalculate dtime */
		if (--togo < 0)
			newdtime();
	      nextart:
#ifdef USE_SCANDIR
	  	/* Take advantage of a prescanned array of articles */
		if (gl->nnums > 0)
			n = gl->nums[gl->ino++];
		else
#endif
		  ++n;
	}

	if (verbose > 1) {
		i = n - gl->no + 1;
		(void)printf("%s: wanted to delete %d article%s\n",
		    prog, i, PLURAL(i));
		i = deleted - olddeleted;
		if (i)
			(void)printf("%s: deleted %d article%s\n",
			    prog, i, PLURAL(i));
	}

	/* Update oldest article and its timestamp */
	gl->no = n;
	if (t != 0)
		gl->to = t;
	else
		gl->to = currenttime;

	return(status);
}

/* Article deletion routine; returns desired exit() status */
static int
dexpire()
{
	register struct class *cl;
	register struct group *gl;
	register time_t cutoff, d;
	register int i, pass, olddeleted;
	register long n, oldfreed_blk;
	int done;

	/* Calculate initial dtime */
	newdtime();

	/*
	 * Basic rules:
	 *
	 *   - We hit the lowest priority classes first and work our
	 *     way up to the highest priority classes
	 *
	 *   - If a pass through all classes and newsgroups doesn't
	 *     yield enough space, we lower the "standard" by some
	 *     amount of time and start over.
	 *
	 *   - We don't test the free space as often when we have a
	 *     long ways to go.
	 *
	 *   - We lower the "standard" time at the start of the loop
	 *     since findstandard() picks a value that won't delete
	 *     any files.
	 */
	pass = 0;
	done = 0;
	for (;;) {
		/* Internal book keeping */
		olddeleted = deleted;
		oldfreed_blk = freed_blk;
		++pass;

		/* Increase standard time */
		standardtime += dtime;
		deltatime = currenttime - standardtime;
		if (deltatime <= 0) {
			(void)fprintf(stderr,
			    "%s: Standard time is in the future!\n", prog);
				return(1);
		}
		if (verbose > 1) {
			(void)printf("%s: Delta time: %s\n",
			    prog, fmtdelta(deltatime));
		}

		/* Proceed from low class to high class */
		for (cl = classlist; cl; cl = cl->next) {
			/* Calculate cutoff for this class */
			d = (deltatime * cl->pri) / standardclass->pri;

			/* Skip entire class if we're reached the minimum */
			if (cl->minimum >= d) {
				if (verbose > 1) {
					(void)printf(
					    "%s: Skip   %3d%%: %s (%s)\n",
					    prog, (100 * cl->pri) /
					    standardclass->pri,
					    fmtdate(currenttime - cl->minimum),
					    fmtdelta(cl->minimum));
				}
				continue;
			}

			/* Adjust for maximum */
			if ((cl->maximum > 0) && (d > cl->maximum)) {
				d = cl->maximum;
			}

			cutoff = currenttime - d;

			/* Update class time */
			cl->to = cutoff;
			if (verbose > 1) {
				(void)printf("%s: Cutoff %3d%%: %s (%s)\n",
				    prog, (100 * cl->pri) / standardclass->pri,
				    fmtdate(cutoff), fmtdelta(d));
			}
			if (cl->maximum > 0)
				d = currenttime - cl->maximum;
			else
				d = 0;

			/* Run through the groups */
			for (gl = cl->group; gl; gl = gl->next)
				if (expireone(gl, cl, cutoff, d)) {
					++done;
					goto bail;
				}
		}
bail:
		i = deleted - olddeleted;
		if ((verbose && i > 0) || verbose > 1) {
			n = BLK2KB(freed_blk - oldfreed_blk);
			(void)printf(
		    "%s: Pass %d %sdeleted %d article%s (%ld Kbyte%s)\n",
			    prog, pass, nflag ? "would have " : "",
			    i, PLURAL(i), n, PLURAL(n));
		}
		if (done)
			return(0);
	}
}

/* Dump out internal data structures */
static void
dump()
{
	register struct class *cl;
	register struct rule *rp;
	register struct group *gl;

	(void)printf("------\nDiagnostic dump:\n");
	(void)printf("\nClasses:\n");
	for (cl = classlist; cl; cl = cl->next) {
		(void)printf("%3d:", cl->pri);
		if (cl->minimum > 0)
			(void)printf(" minimum %s", fmtdelta(cl->minimum));
		if (cl->maximum > 0)
			(void)printf(" maximum %s", fmtdelta(cl->maximum));
		if (cl == standardclass) {
			if (cl->minimum > 0)
				putchar(',');
			(void)printf(" standard");
		}
		putchar('\n');
	}

	printf("\nRules:\n");
	for (rp = rulelist; rp; rp = rp->next) {
		(void)printf("%-32s %c\t",
		    rp->ngn, rp->mod);
		if (rp->class == 0)
			(void)printf("<never>");
		else {
			(void)printf("%d", rp->class->pri);
			if (rp->class->minimum)
				(void)printf("\tmin %s",
				    fmtdelta(rp->class->minimum));
			if (rp->class->maximum)
				(void)printf("\tmax %s",
				    fmtdelta(rp->class->maximum));
		}
		putchar('\n');
	}

	(void)printf("\nGroups:\n");
	for (cl = classlist; cl; cl = cl->next) {
		(void)printf("%3d:", cl->pri);
		if (cl->group == 0) {
			putchar('\n');
			continue;
		}
		for (gl = cl->group; gl; gl = gl->next) {
			(void)printf("\t%-32s [%ld, %ld]",
			    gl->ngn, gl->no, gl->nn);
			if (gl == cl->group && cl->minimum > 0)
				(void)printf(" (min %s)", fmtdelta(cl->minimum));
			if (gl == cl->group && cl->maximum > 0)
				(void)printf(" (max %s)", fmtdelta(cl->maximum));
			putchar('\n');
		}
	}
	(void)printf("------\n");
}

static void
pfcleanup()
{
    (void)unlink(pidfile);
}

static void
sighand(s)
    int s;
{
    (void)fflush(stderr);
    (void)fprintf(stderr, "%s: Killed by signal %d\n", prog, s);
    exit(255);
}

static void
set_signals(s)
    void (*s)();
{
    (void)signal(SIGINT, s);
    (void)signal(SIGTERM, s);
#ifdef SIGXCPU
    (void)signal(SIGXCPU, s);
#endif
}
